from __future__ import annotations

from typing import Any

import pandas as pd

from ..core.task import DataLocation
from ..pandas import DataFrame as _DataFrame
from .file import load_file


class DataFrame(_DataFrame):
    def __init__(
        self,
        dataset: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name="",
            location=DataLocation.CLIENT,
            src=None,
            dataset=dataset,
            default=None,
            **kwargs,
        )


def load_dataframe(filename: str) -> pd.DataFrame:
    res = load_file(filename)
    if isinstance(res, pd.DataFrame):
        return res
    else:
        raise ValueError(f"{filename} is not a dataframe")



from __future__ import annotations

from abc import abstractmethod
from typing import List, Tuple, Dict

from ..core.strategy import AnalyticsStrategy, CURVE_TYPE
from ..core.task import GraphNode, InputGraphNode
from .task import HorizontalTask
import delta.dataset


class HorizontalAnalytics(HorizontalTask):
    def __init__(
        self,
        name: str,
        min_clients: int = 2,
        max_clients: int = 2,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
    ) -> None:
        strategy = AnalyticsStrategy(
            min_clients=min_clients,
            max_clients=max_clients,
            wait_timeout=wait_timeout,
            connection_timeout=connection_timeout,
            precision=precision,
            curve=curve,
        )
        super().__init__(name, strategy)

    @abstractmethod
    def execute(self, **inputs: InputGraphNode) -> GraphNode | Tuple[GraphNode]:
        ...

    @abstractmethod
    def dataset(self) -> Dict[str, delta.dataset.DataFrame]:
        ...

    def _build_graph(self) -> Tuple[List[InputGraphNode], List[GraphNode]]:
        inputs = self.dataset()
        for name, node in inputs.items():
            node.name = name
        outputs = self.execute(**inputs)
        if isinstance(outputs, tuple):
            return list(inputs.values()), list(outputs)
        else:
            return list(inputs.values()), [outputs]



import os
from os import PathLike
from typing import IO, Union

import cloudpickle as pickle

from ..core.task import Task


def dump_task(file: Union[str, PathLike, IO[bytes]], task: Task):
    if isinstance(file, str):
        with open(file, mode="wb") as f:
            pickle.dump(task, f)
    elif isinstance(file, PathLike):
        filename = os.fspath(file)
        with open(filename, mode="wb") as f:
            pickle.dump(task, f)
    else:
        return pickle.dump(task, file)


def load_task(file: Union[str, PathLike, IO[bytes]]) -> Task:
    if isinstance(file, PathLike):
        filename = os.fspath(file)
        with open(filename, mode="rb") as f:
            res = pickle.load(f)
    elif isinstance(file, str):
        with open(file, mode="rb") as f:
            res = pickle.load(f)
    else:
        res = pickle.load(file)
    assert isinstance(res, Task), "File is not a task file"
    return res



from __future__ import annotations

from typing import Dict, List

import numpy as np
import torch

from .merge import EpochMergeStrategy, MergeStrategy
from .result import ResultStrategy, WeightResultStrategy
from .select import RandomSelectStrategy, SelectStrategy
from .strategy import CURVE_TYPE, Strategy


class LearningStrategy(Strategy):
    def __init__(
        self,
        name: str,
        select_strategy: SelectStrategy,
        merge_strategy: MergeStrategy,
        result_strategy: ResultStrategy,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        fault_tolerant: bool = False,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
    ) -> None:
        super().__init__(
            name=name,
            select_strategy=select_strategy,
            wait_timeout=wait_timeout,
            connection_timeout=connection_timeout,
            fault_tolerant=fault_tolerant,
            precision=precision,
            curve=curve,
        )
        self.merge_strategy = merge_strategy
        self.result_strategy = result_strategy

    def should_merge(self, epoch: int, iteration: int, epoch_finished: bool) -> bool:
        return self.merge_strategy.should_merge(epoch, iteration, epoch_finished)

    def params_to_weight(self, params: Dict[str, torch.Tensor]) -> np.ndarray:
        return self.result_strategy.params_to_weight(params)

    def weight_to_params(self, weight: np.ndarray, params: Dict[str, torch.Tensor]):
        return self.result_strategy.weight_to_params(weight, params)

    def params_to_result(
        self, params: Dict[str, torch.Tensor], last_weight: np.ndarray
    ) -> np.ndarray:
        return self.result_strategy.params_to_result(params, last_weight)

    def result_to_params(self, result: np.ndarray, params: Dict[str, torch.Tensor]):
        return self.result_strategy.result_to_params(result, params)



from __future__ import annotations

import os
from typing import Any

import numpy as np
import pandas as pd
import torch
from PIL import Image, UnidentifiedImageError


class UnsupportedFileError(Exception):
    def __init__(self, filename: str) -> None:
        self.filename = filename

    def __str__(self) -> str:
        return f"unsupported file {self.filename}"


def load_file(
    filename: str, **kwargs: Any
) -> np.ndarray | torch.Tensor | pd.DataFrame | Image.Image:
    if not os.path.exists(filename):
        raise ValueError(f"{filename} does not exist")

    if filename.endswith(".npz"):
        result = np.load(filename)["arr_0"]
    elif filename.endswith(".npy"):
        result = np.load(filename)
    elif filename.endswith(".pt"):
        result = torch.load(filename)
    elif filename.endswith(".csv"):
        if "sep" in kwargs:
            kwargs.pop("sep")
        result = pd.read_csv(filename, sep=",", **kwargs)
    elif filename.endswith(".tsv") or filename.endswith(".txt"):
        if "sep" in kwargs:
            kwargs.pop("sep")
        result = pd.read_csv(filename, sep=r"\s+", **kwargs)
    elif filename.endswith(".xls") or filename.endswith(".xlsx"):
        result = pd.read_excel(filename, **kwargs)
    else:
        try:
            result = Image.open(filename, **kwargs)
        except UnidentifiedImageError:
            raise UnsupportedFileError(filename)
    return result


def load_dataframe(filename: str) -> pd.DataFrame:
    res = load_file(filename)
    if isinstance(res, pd.DataFrame):
        return res
    else:
        raise ValueError(f"{filename} is not a dataframe")


import os
from os import PathLike
from typing import IO, Union

import numpy as np


def load_arr(file: Union[str, PathLike, IO[bytes]]) -> np.ndarray:
    arr = np.load(file)["arr_0"]
    return arr


def dump_arr(file: Union[str, PathLike, IO[bytes]], arr: np.ndarray):
    if isinstance(file, str):
        with open(file, mode="wb") as f:
            np.savez_compressed(f, arr)
    elif isinstance(file, PathLike):
        filename = os.fspath(file)
        with open(filename, mode="wb") as f:
            np.savez_compressed(f, arr)
    else:
        np.savez_compressed(file, arr)



from tempfile import TemporaryFile

import httpx

from . import serialize
from .core.task import Task


class DeltaNode(object):
    def __init__(self, url: str) -> None:
        self._url = url

    def create_task(self, task: Task) -> int:
        url = f"{self._url}/v1/task"
        with TemporaryFile(mode="w+b") as file:
            serialize.dump_task(file, task)
            file.seek(0)
            resp = httpx.post(url, files={"file": file})
            resp.raise_for_status()
            data = resp.json()
            task_id = data["task_id"]
            return task_id



from __future__ import annotations

import random
from abc import ABC, abstractmethod
from typing import List


class SelectStrategy(ABC):
    def __init__(self, min_clients: int, max_clients: int) -> None:
        self.min_clients = min_clients
        self.max_clients = max_clients

    @abstractmethod
    def select(
        self, candidates: List[str], last_partners: List[str] | None = None
    ) -> List[str]:
        ...


class RandomSelectStrategy(SelectStrategy):
    def __init__(self, min_clients: int, max_clients: int) -> None:
        super().__init__(min_clients, max_clients)
        self.max_clients = max_clients

    def select(
        self, candidates: List[str], last_partners: List[str] | None = None
    ) -> List[str]:
        if len(candidates) < self.min_clients:
            raise ValueError("not enough clients in select candidates")
        res = candidates
        if len(res) > self.max_clients:
            res = random.sample(res, k=self.max_clients)
        return res


class SameSelectStrategy(SelectStrategy):
    def __init__(self, min_clients: int, max_clients: int) -> None:
        super().__init__(min_clients, max_clients)
        self.max_clients = max_clients

    def select(
        self, candidates: List[str], last_partners: List[str] | None = None
    ) -> List[str]:
        if len(candidates) < self.min_clients:
            raise ValueError("Not enough clients in select candidates")

        if last_partners is None:
            res = candidates
            if len(res) > self.max_clients:
                res = random.sample(res, k=self.max_clients)
            return res
        else:
            assert (
                self.min_clients <= len(last_partners) <= self.max_clients
            ), "Last partners length is invalid"
            return last_partners



import operator
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, TypeVar


def local_op(name: str):
    def _op(left, right, **kwargs):
        return getattr(left, name)(right, **kwargs)

    return _op


def rlocal_op(name: str):
    def _op(left, right, **kwargs):
        return getattr(right, name)(left, **kwargs)

    return _op


T = TypeVar("T")

class OpMixin(ABC, Generic[T]):
    @abstractmethod
    def _arith_method(self, other, name: str, op: Callable[[Any, Any], Any]) -> T:
        ...

    def __add__(self, other) -> T:
        return self._arith_method(other, "__add__", operator.add)

    def __radd__(self, other) -> T:
        def radd(left, right):
            return right + left

        return self._arith_method(other, "__radd__", radd)

    def __sub__(self, other) -> T:
        return self._arith_method(other, "__sub__", operator.sub)

    def __rsub__(self, other) -> T:
        def rsub(left, right):
            return right - left

        return self._arith_method(other, "__rsub__", rsub)

    def __mul__(self, other) -> T:
        return self._arith_method(other, "__mul__", operator.mul)

    def __rmul__(self, other) -> T:
        def rmul(left, right):
            return right * left

        return self._arith_method(other, "__rmul__", rmul)

    def __truediv__(self, other) -> T:
        return self._arith_method(other, "__truediv__", operator.truediv)

    def __rtruediv__(self, other) -> T:
        def rtruediv(left, right):
            return right / left

        return self._arith_method(other, "__rtruediv__", rtruediv)

    def __floordiv__(self, other) -> T:
        return self._arith_method(other, "__floordiv__", operator.floordiv)

    def __rfloordiv__(self, other) -> T:
        def rfloordiv(left, right):
            return right // left

        return self._arith_method(other, "__rfloordiv__", rfloordiv)

    def __mod__(self, other) -> T:
        return self._arith_method(other, "__mod__", operator.mod)

    def __rmod__(self, other) -> T:
        def rmod(left, right):
            if isinstance(right, str):
                typ = type(left).__name__
                raise TypeError(f"{typ} cannot perform the operation mod")

            return right % left

        return self._arith_method(other, "__rmod__", rmod)

    def __divmod__(self, other) -> T:
        return self._arith_method(other, "divmod", divmod)

    def __rdivmod__(self, other) -> T:
        def rdivmod(left, right):
            return divmod(right, left)

        return self._arith_method(other, "__rdivmod__", rdivmod)

    def __pow__(self, other) -> T:
        return self._arith_method(other, "__pow__", operator.pow)

    def __rpow__(self, other) -> T:
        def rpow(left, right):
            return right**left

        return self._arith_method(other, "__rpow__", rpow)



from .analytics import HorizontalAnalytics
from .learning import HorizontalLearning, FaultTolerantFedAvg, FedAvg


__all__ = ["HorizontalAnalytics", "HorizontalLearning", "FaultTolerantFedAvg", "FedAvg"]


from .select import SameSelectStrategy
from .strategy import CURVE_TYPE, Strategy


class AnalyticsStrategy(Strategy):
    def __init__(
        self,
        min_clients: int = 2,
        max_clients: int = 2,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
    ) -> None:
        super().__init__(
            name="analytics",
            select_strategy=SameSelectStrategy(min_clients, max_clients),
            wait_timeout=wait_timeout,
            connection_timeout=connection_timeout,
            fault_tolerant=False,
            precision=precision,
            curve=curve,
        )



from __future__ import annotations

from typing import Any, Dict, List, Tuple

from .dataset import load_dataframe, load_dataset
from .core.task import (ClientContext, DataFormat, DataLocation, DataNode,
                        InputGraphNode, ServerContext, Task)


class Aggregator(object):
    def __init__(self) -> None:
        self._data = {}

    def upload(self, name: str, data: Dict[str, Any]):
        self._data[name] = data

    def gather(self, name: str) -> Tuple[Dict[str, Any], int]:
        return self._data[name], 1


class DebugClientContext(ClientContext):
    def __init__(self, aggregator: Aggregator, server_ctx: ServerContext) -> None:
        self._state = {}
        self._aggregator = aggregator
        self._server_ctx = server_ctx

    def get(self, *vars: DataNode) -> List[Any]:
        res = []
        for var in vars:
            value = None
            if var.location == DataLocation.CLIENT:
                value = self._state.get(var.name)
                if value is None and isinstance(var, InputGraphNode):
                    if var.filename is not None and var.format is not None:
                        value = self.load_data(var.filename, var.format, **var.kwargs)
                    elif var.default is not None:
                        value = var.default

                    if value is not None:
                        self._state[var.name] = value
            else:
                value = self._server_ctx.get(var)[0]

            if value is None:
                raise ValueError(f"Cannot get var {var.name}")

            res.append(value)

        return res

    def set(self, *pairs: Tuple[DataNode, Any]):
        for var, data in pairs:
            self._state[var.name] = data

    def has(self, key: str) -> bool:
        return key in self._state

    def upload(self, name: str, data: Dict[str, Any]):
        self._aggregator.upload(name, data)

    def load_data(self, filename: str, format: DataFormat, **kwargs):
        if format == DataFormat.DATASET:
            return load_dataset(filename, **kwargs)
        elif format == DataFormat.DATAFRAME:
            return load_dataframe(filename)

    def clear(self):
        self._state.clear()


class DebugServerContext(ServerContext):
    def __init__(self, aggregator: Aggregator) -> None:
        self._state = {}
        self._aggregator = aggregator

    def get(self, *vars: DataNode) -> List[Any]:
        res = []
        for var in vars:
            value = self._state.get(var.name)
            if value is None and isinstance(var, InputGraphNode):
                if var.default is not None:
                    value = var.default
                    self._state[var.name] = value

            if value is None:
                raise ValueError(f"Cannot get var {var.name}")

            res.append(value)

        return res

    def set(self, *pairs: Tuple[DataNode, Any]):
        for var, data in pairs:
            self._state[var.name] = data

    def has(self, key: str) -> bool:
        return key in self._state

    def gather(self, name: str) -> Tuple[Dict[str, Any], int]:
        return self._aggregator.gather(name)

    def clear(self):
        self._state.clear()


def debug(task: Task, **data: Any):
    aggregator = Aggregator()
    server_context = DebugServerContext(aggregator)
    client_context = DebugClientContext(aggregator, server_context)

    for var in task.inputs:
        if var.name in data:
            if var.location == DataLocation.CLIENT:
                client_context.set((var, data[var.name]))
            elif var.location == DataLocation.SERVER:
                server_context.set((var, data[var.name]))

    for step in task.steps:
        step.map(client_context)
        step.reduce(server_context)

    res = tuple(server_context.get(*task.outputs))
    if len(res) == 1:
        return res[0]
    else:
        return res



from .dataframe import load_dataframe, DataFrame
from .dataset import load_dataset, split_dataset, Dataset
from .file import load_file

__all__ = [
    "load_dataset",
    "split_dataset",
    "load_file",
    "load_dataframe",
    "DataFrame",
    "Dataset",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, List

import numpy as np
import torch


class ResultStrategy(ABC):
    @abstractmethod
    def weight_to_params(self, weight: np.ndarray, params: Dict[str, torch.Tensor]):
        ...

    @abstractmethod
    def params_to_weight(self, params: Dict[str, torch.Tensor]) -> np.ndarray:
        ...

    @abstractmethod
    def params_to_result(
        self, params: Dict[str, torch.Tensor], last_weight: np.ndarray
    ) -> np.ndarray:
        ...

    @abstractmethod
    def result_to_params(
        self, result: np.ndarray, params: Dict[str, torch.Tensor]
    ):
        ...


class WeightResultStrategy(ResultStrategy):
    def params_to_weight(self, params: Dict[str, torch.Tensor]) -> np.ndarray:
        arrs = [p.detach().cpu().ravel().numpy() for p in params.values()]
        result = np.concatenate(arrs, axis=0)
        return result

    def weight_to_params(self, weight: np.ndarray, params: Dict[str, torch.Tensor]):
        offset = 0
        with torch.no_grad():
            for p in params.values():
                numel = p.numel()
                weight_slice = weight[offset : offset + numel]
                offset += numel
                weight_tensor = (
                    torch.from_numpy(weight_slice)
                    .to(p.dtype)
                    .to(p.device)
                    .resize_(p.shape)
                )
                p.copy_(weight_tensor)

    def params_to_result(self, params: Dict[str, torch.Tensor], last_weight: np.ndarray) -> np.ndarray:
        return self.params_to_weight(params)
    
    def result_to_params(self, result: np.ndarray, params: Dict[str, torch.Tensor]):
        return self.weight_to_params(result, params)



from .analytics import AnalyticsStrategy
from .learning import LearningStrategy
from .merge import EpochMergeStrategy, IterMergeStrategy, MergeStrategy
from .result import ResultStrategy, WeightResultStrategy
from .select import RandomSelectStrategy, SameSelectStrategy, SelectStrategy
from .strategy import CURVE_TYPE, Strategy

__all__ = [
    "AnalyticsStrategy",
    "LearningStrategy",
    "EpochMergeStrategy",
    "IterMergeStrategy",
    "MergeStrategy",
    "ResultStrategy",
    "WeightResultStrategy",
    "RandomSelectStrategy",
    "SameSelectStrategy",
    "SelectStrategy",
    "CURVE_TYPE",
    "Strategy",
]



from __future__ import annotations

from typing import List, Literal

from .select import SelectStrategy

CURVE_TYPE = Literal[
    "secp192r1", "secp224r1", "secp256k1", "secp256r1", "secp384r1", "secp521r1"
]


class Strategy(object):
    def __init__(
        self,
        name: str,
        select_strategy: SelectStrategy,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        fault_tolerant: bool = False,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
    ) -> None:
        self.name = name
        self.select_strategy = select_strategy
        self.wait_timeout = wait_timeout
        self.connection_timeout = connection_timeout
        self.fault_tolerant = fault_tolerant
        self.precision = precision
        self.curve = curve

    def select(
        self, candidates: List[str], last_partners: List[str] | None = None
    ) -> List[str]:
        return self.select_strategy.select(candidates, last_partners)



from __future__ import annotations

from abc import ABC, abstractmethod


class MergeStrategy(ABC):
    @abstractmethod
    def should_merge(self, epoch: int, iteration: int, epoch_finished: bool) -> bool:
        ...


class EpochMergeStrategy(MergeStrategy):
    def __init__(self, epoch: int) -> None:
        assert epoch > 0, "Epoch should be greater than 0"
        self.epoch = epoch

    def should_merge(self, epoch: int, iteration: int, epoch_finished: bool) -> bool:
        return epoch_finished and epoch % self.epoch == 0


class IterMergeStrategy(MergeStrategy):
    def __init__(self, iteration: int) -> None:
        assert iteration > 0, "Iteration should be greater than 0"
        self.iteration = iteration

    def should_merge(self, epoch: int, iteration: int, epoch_finished: bool) -> bool:
        return (not epoch_finished) and iteration % self.iteration == 0



from __future__ import annotations

from typing import Any, Callable, Dict, List, Literal, Type

import pandas

from delta.core.task import (
    DataFormat,
    DataLocation,
    DataNode,
    GraphNode,
    InputGraphNode,
    MapOperator,
    MapReduceOperator,
    Operator,
    ReduceOperator,
)

from .series import Series
from .op_mixin import OpMixin, local_op, rlocal_op

AxisType = Literal[0, 1]


class DataFrame(InputGraphNode, OpMixin["DataFrame"]):
    def __init__(
        self,
        name: str = "",
        location: DataLocation = DataLocation.CLIENT,
        src: Operator | None = None,
        dataset: str | None = None,
        default: Any | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name, location, src, DataFormat.DATAFRAME, dataset, default, **kwargs
        )

    def _dispatch_map_reduce(
        self,
        name: str,
        axis: AxisType,
        map_reduce_op: MapReduceOperator,
        map_op: MapOperator | None = None,
        reduce_op: ReduceOperator | None = None,
        reduce_axis: AxisType = 0,
        **kwargs,
    ):
        output = Series(location=DataLocation.CLIENT)

        if self.location == DataLocation.SERVER:
            if reduce_op is None:

                class _Reduce(ReduceOperator):
                    def reduce(self, data: pandas.DataFrame) -> pandas.Series:
                        return getattr(data, self.name)(**self.kwargs)

                reduce_op = _Reduce(
                    name=name, inputs=[self], outputs=[output], **kwargs
                )
            else:
                assert len(reduce_op.outputs) == 0
                reduce_op.outputs.append(output)
            output.src = reduce_op
            output.location = DataLocation.SERVER

        elif axis == reduce_axis:
            assert (
                len(map_reduce_op.reduce_outputs) == 0
                and len(map_reduce_op.outputs) == 0
            )
            map_reduce_op.reduce_outputs.append(output)
            map_reduce_op.outputs.append(output)
            output.src = map_reduce_op
            output.location = DataLocation.SERVER

        else:
            if map_op is None:

                class _Map(MapOperator):
                    def map(self, data: pandas.DataFrame) -> pandas.Series:
                        return getattr(data, self.name)(**self.kwargs)

                map_op = _Map(name=name, inputs=[self], outputs=[output], **kwargs)
            else:
                assert len(map_op.outputs) == 0
                map_op.outputs.append(output)
            output.src = map_op

        return output

    def _dispatch_binary_op(
        self,
        other: "DataFrame" | "Series" | List[float] | float,
        op_name: str,
        op: Callable[..., Any],
        **kwargs: Any,
    ) -> "DataFrame":
        if not isinstance(other, GraphNode):
            other_node = InputGraphNode(default=other, location=DataLocation.SERVER)
        else:
            other_node = other

        if (
            self.location == DataLocation.CLIENT
            or other_node.location == DataLocation.CLIENT
        ):
            location = DataLocation.CLIENT
        else:
            location = DataLocation.SERVER

        output = DataFrame(location=location)

        if location == DataLocation.SERVER:

            class _Reduce(ReduceOperator):
                def __init__(
                    self,
                    name: str,
                    inputs: List[DataNode],
                    outputs: List[DataNode],
                    op: Callable[..., Any],
                    **kwargs: Any,
                ) -> None:
                    super().__init__(name, inputs, outputs, **kwargs)
                    self.op = op

                def reduce(
                    self,
                    left: pandas.DataFrame,
                    right: pandas.DataFrame | pandas.Series | List[float] | float,
                ):
                    return self.op(left, right, **self.kwargs)

            _op = _Reduce(
                op_name, inputs=[self, other_node], outputs=[output], op=op, **kwargs
            )
            output.src = _op

        else:

            class _Map(MapOperator):
                def __init__(
                    self,
                    name: str,
                    inputs: List[DataNode],
                    outputs: List[DataNode],
                    op: Callable[..., Any],
                    **kwargs: Any,
                ) -> None:
                    super().__init__(name, inputs, outputs, **kwargs)
                    self.op = op

                def map(
                    self,
                    left: pandas.DataFrame,
                    right: pandas.DataFrame | pandas.Series | List[float] | float,
                ):
                    return self.op(left, right, **self.kwargs)

            _op = _Map(
                op_name, inputs=[self, other_node], outputs=[output], op=op, **kwargs
            )
            output.src = _op

        return output

    def all(
        self,
        axis: AxisType = 0,
        bool_only: bool = False,
        skipna: bool = True,
    ) -> "Series":
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType = 0,
                bool_only: bool = False,
                skipna: bool = True,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    bool_only=bool_only,
                    skipna=skipna,
                )
                self.axis: AxisType = axis
                self.bool_only = bool_only
                self.skipna = skipna

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                res = data.all(
                    axis=self.axis, bool_only=self.bool_only, skipna=self.skipna
                )
                return {"all": res}

            def reduce(
                self, results: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                agg_all = results.get("all")
                if agg_all is not None:
                    return agg_all == node_count
                else:
                    raise ValueError("aggregate result miss key all")

        map_reduce_op = _MapReduce(
            "all",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            bool_only=bool_only,
            skipna=skipna,
        )

        return self._dispatch_map_reduce(
            name="all",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            bool_only=bool_only,
            skipna=skipna,
        )

    def any(
        self,
        axis: AxisType = 0,
        bool_only: bool = False,
        skipna: bool = True,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType = 0,
                bool_only: bool = False,
                skipna: bool = True,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    bool_only=bool_only,
                    skipna=skipna,
                )
                self.axis: AxisType = axis
                self.bool_only = bool_only
                self.skipna = skipna

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                res = data.any(
                    axis=self.axis, bool_only=self.bool_only, skipna=self.skipna
                )
                return {"all": res}

            def reduce(
                self, results: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                agg_all = results.get("all")
                if agg_all is not None:
                    return agg_all > 0
                else:
                    raise ValueError("aggregate result miss key all")

        map_reduce_op = _MapReduce(
            "any",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            bool_only=bool_only,
            skipna=skipna,
        )

        return self._dispatch_map_reduce(
            name="any",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            bool_only=bool_only,
            skipna=skipna,
        )

    def count(
        self,
        axis: AxisType = 0,
        numeric_only: bool = False,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                numeric_only: bool,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    numeric_only=numeric_only,
                )
                self.axis: AxisType = axis
                self.numeric_only = numeric_only

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                res = data.count(axis=self.axis, numeric_only=self.numeric_only)
                return {"count": res}

            def reduce(
                self, data: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                count = data.get("count")
                if count is not None:
                    return count
                else:
                    raise ValueError("aggregate result miss key count")

        map_reduce_op = _MapReduce(
            name="count",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            numeric_only=numeric_only,
        )

        return self._dispatch_map_reduce(
            name="count",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            numeric_only=numeric_only,
        )

    def sum(
        self,
        axis: AxisType = 0,
        skipna: bool = True,
        numeric_only: bool = False,
        min_count: int = 0,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                skipna: bool,
                numeric_only: bool,
                min_count: int,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    skipna=skipna,
                    numeric_only=numeric_only,
                    min_count=min_count,
                )
                self.axis: AxisType = axis
                self.skipna = skipna
                self.numeric_only = numeric_only
                self.min_count = min_count

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                res = data.sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                    min_count=self.min_count,
                )
                return {"sum": res}

            def reduce(
                self, data: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                res = data.get("sum")
                if res is not None:
                    return res
                else:
                    raise ValueError("aggregate result miss key sum")

        map_reduce_op = _MapReduce(
            name="sum",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            skipna=skipna,
            numeric_only=numeric_only,
            min_count=min_count,
        )

        return self._dispatch_map_reduce(
            name="sum",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            skipna=skipna,
            numeric_only=numeric_only,
            min_count=min_count,
        )

    def mean(
        self,
        axis: AxisType = 0,
        skipna: bool = True,
        numeric_only: bool = False,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                skipna: bool,
                numeric_only: bool,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    skipna=skipna,
                    numeric_only=numeric_only,
                )
                self.axis: AxisType = axis
                self.skipna = skipna
                self.numeric_only = numeric_only

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                sum = data.sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                count = data.count(axis=self.axis, numeric_only=self.numeric_only)
                if not self.skipna:
                    count += data.isna().sum(axis=self.axis)
                return {"sum": sum, "count": count}

            def reduce(
                self, data: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                sum = data.get("sum")
                if sum is None:
                    raise ValueError("aggregate result miss key sum")
                count = data.get("count")
                if count is None:
                    raise ValueError("aggregate result miss key count")

                return sum / count

        map_reduce_op = _MapReduce(
            name="mean",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            skipna=skipna,
            numeric_only=numeric_only,
        )

        return self._dispatch_map_reduce(
            name="mean",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            skipna=skipna,
            numeric_only=numeric_only,
        )

    def std(
        self,
        axis: AxisType = 0,
        skipna: bool = True,
        ddof: int = 1,
        numeric_only: bool = False,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                skipna: bool,
                ddof: int,
                numeric_only: bool,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    skipna=skipna,
                    ddof=ddof,
                    numeric_only=numeric_only,
                )
                self.axis: AxisType = axis
                self.skipna = skipna
                self.ddof = ddof
                self.numeric_only = numeric_only

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                sum1 = data.sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                sum2 = data.pow(2).sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                count = data.count(
                    axis=self.axis,
                    numeric_only=self.numeric_only,
                )
                if not self.skipna:
                    count += data.isna().sum(axis=self.axis)
                return {"sum1": sum1, "sum2": sum2, "count": count}

            def reduce(
                self, data: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                sum1 = data.get("sum1")
                if sum1 is None:
                    raise ValueError("aggregate result miss key sum1")
                sum2 = data.get("sum2")
                if sum2 is None:
                    raise ValueError("aggregate result miss key sum2")
                count = data.get("count")
                if count is None:
                    raise ValueError("aggregate result miss key count")

                res = (sum2 / count - (sum1 / count) ** 2) * count / (count - self.ddof)
                return res ** (1 / 2)

        map_reduce_op = _MapReduce(
            name="std",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

        return self._dispatch_map_reduce(
            name="std",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

    def var(
        self,
        axis: AxisType = 0,
        skipna: bool = True,
        ddof: int = 1,
        numeric_only: bool = False,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                skipna: bool,
                ddof: int,
                numeric_only: bool,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    skipna=skipna,
                    ddof=ddof,
                    numeric_only=numeric_only,
                )
                self.axis: AxisType = axis
                self.skipna = skipna
                self.ddof = ddof
                self.numeric_only = numeric_only

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                sum1 = data.sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                sum2 = data.pow(2).sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                count = data.count(
                    axis=self.axis,
                    numeric_only=self.numeric_only,
                )
                if not self.skipna:
                    count += data.isna().sum(axis=self.axis)
                return {"sum1": sum1, "sum2": sum2, "count": count}

            def reduce(
                self, data: Dict[str, pandas.Series], node_count: int
            ) -> pandas.Series:
                sum1 = data.get("sum1")
                if sum1 is None:
                    raise ValueError("aggregate result miss key sum1")
                sum2 = data.get("sum2")
                if sum2 is None:
                    raise ValueError("aggregate result miss key sum2")
                count = data.get("count")
                if count is None:
                    raise ValueError("aggregate result miss key count")

                res = (sum2 / count - (sum1 / count) ** 2) * count / (count - self.ddof)
                return res

        map_reduce_op = _MapReduce(
            name="var",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

        return self._dispatch_map_reduce(
            name="var",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

    def sem(
        self,
        axis: AxisType = 0,
        skipna: bool = True,
        ddof: int = 1,
        numeric_only: bool = False,
    ):
        class _MapReduce(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                axis: AxisType,
                skipna: bool,
                ddof: int,
                numeric_only: bool,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    axis=axis,
                    skipna=skipna,
                    ddof=ddof,
                    numeric_only=numeric_only
                )
                self.axis: AxisType = axis
                self.skipna = skipna
                self.ddof = ddof
                self.numeric_only = numeric_only

            def map(self, data: pandas.DataFrame) -> Dict[str, pandas.Series]:
                sum1 = data.sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                sum2 = data.pow(2).sum(
                    axis=self.axis,
                    skipna=self.skipna,
                    numeric_only=self.numeric_only,
                )
                count = data.count(
                    axis=self.axis,
                    numeric_only=self.numeric_only,
                )
                if not self.skipna:
                    count += data.isna().sum(axis=self.axis)
                return {"sum1": sum1, "sum2": sum2, "count": count}

            def reduce(self, data: Dict[str, pandas.Series], node_count: int) -> pandas.Series:
                sum1 = data.get("sum1")
                if sum1 is None:
                    raise ValueError("aggregate result miss key sum1")
                sum2 = data.get("sum2")
                if sum2 is None:
                    raise ValueError("aggregate result miss key sum2")
                count = data.get("count")
                if count is None:
                    raise ValueError("aggregate result miss key count")

                res = ((sum2 / count - (sum1 / count) ** 2) / (count - self.ddof)) ** (1 / 2)
                return res

        map_reduce_op = _MapReduce(
            name="sem",
            map_inputs=[self],
            reduce_inputs=[],
            map_outputs=[],
            reduce_outputs=[],
            axis=axis,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

        return self._dispatch_map_reduce(
            name="sem",
            axis=axis,
            map_reduce_op=map_reduce_op,
            reduce_axis=0,
            skipna=skipna,
            ddof=ddof,
            numeric_only=numeric_only,
        )

    def _arith_method(
        self, other, name: str, op: Callable[[Any, Any], Any]
    ) -> "DataFrame":
        return self._dispatch_binary_op(other, name, op)

    def add(
        self, other: "DataFrame", axis: AxisType = 1, fill_value: float | None = None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "add", local_op("add"), axis=axis, fill_value=fill_value
        )

    def sub(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "sub", local_op("sub"), axis=axis, fill_value=fill_value
        )

    def mul(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "mul", local_op("mul"), axis=axis, fill_value=fill_value
        )

    def div(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "div", local_op("div"), axis=axis, fill_value=fill_value
        )

    def truediv(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "truediv", local_op("truediv"), axis=axis, fill_value=fill_value
        )

    def floordiv(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "floordiv", local_op("floordiv"), axis=axis, fill_value=fill_value
        )

    def mod(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "mod", local_op("mod"), axis=axis, fill_value=fill_value
        )

    def pow(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "pow", local_op("pow"), axis=axis, fill_value=fill_value
        )

    def radd(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "radd", rlocal_op("radd"), axis=axis, fill_value=fill_value
        )

    def rsub(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rsub", rlocal_op("rsub"), axis=axis, fill_value=fill_value
        )

    def rmul(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rmul", rlocal_op("rmul"), axis=axis, fill_value=fill_value
        )

    def rdiv(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rdiv", rlocal_op("rdiv"), axis=axis, fill_value=fill_value
        )

    def rtruediv(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rtruediv", rlocal_op("rtruediv"), axis=axis, fill_value=fill_value
        )

    def rfloordiv(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other,
            "rfloordiv",
            rlocal_op("rfloordiv"),
            axis=axis,
            fill_value=fill_value,
        )

    def rmod(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rmod", rlocal_op("rmod"), axis=axis, fill_value=fill_value
        )

    def rpow(
        self, other: "DataFrame", axis: AxisType = 1, fill_value=None
    ) -> "DataFrame":
        return self._dispatch_binary_op(
            other, "rpow", rlocal_op("rpow"), axis=axis, fill_value=fill_value
        )



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
import pandas

from .strategy import Strategy

__all__ = [
    "DataNode",
    "OpType",
    "GraphNode",
    "Operator",
    "OperatorGroup",
    "MapOperator",
    "MapOperatorGroup",
    "MapReduceOperator",
    "MapReduceOperatorGroup",
    "ReduceOperator",
    "ReduceOperatorGroup",
    "Context",
    "ClientContext",
    "ServerContext",
    "Step",
    "Task",
    "build",
    "DataLocation",
    "DataFormat",
    "InputGraphNode",
    "TaskConstructer",
    "AggValueType",
    "AggResultType",
]


class DataLocation(str, Enum):
    CLIENT = "CLIENT"
    SERVER = "SERVER"


class DataFormat(str, Enum):
    DATASET = "DATASET"
    DATAFRAME = "DATAFRAME"
    SERIES = "SERIES"


class DataNode(object):
    def __init__(
        self,
        name: str = "",
        location: DataLocation = DataLocation.CLIENT,
    ) -> None:
        self.name = name
        self.location = location

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"


class OpType(str, Enum):
    MAP = "MAP"
    MAP_REDUCE = "MAP_REDUCE"
    REDUCE = "REDUCE"


class Operator(abc.ABC):
    priority: int = 100

    def __init__(
        self, name: str, inputs: List[DataNode], outputs: List[DataNode], **kwargs: Any
    ) -> None:
        """
        A single operator

        name: operator name
        inputs: operator input variables
        outputs: operator output variables
        kwargs: key word arguments for the operator implementation
        """
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.kwargs = kwargs

    @property
    @abc.abstractmethod
    def type(self) -> OpType:
        ...

    def __repr__(self) -> str:
        return f"{self.name}(type={self.type}, inputs={self.inputs}, outputs={self.outputs}, kwargs={self.kwargs})"

    def __str__(self) -> str:
        return f"""
            {self.name}
            (
                type={self.type}, 
                inputs={self.inputs}, 
                outputs={self.outputs}, 
                kwargs={self.kwargs}
            )
            """

    def __lt__(self, other: "Operator"):
        return self.priority < other.priority


class OperatorGroup(object):
    def __init__(self, *ops: Operator) -> None:
        """
        A group of operators which are independent with each other.
        Can execute these operators parallelly.

        ops: operators
        """
        self.ops = sorted(ops)

        input_set = set()
        output_set = set()
        for op in ops:
            input_set = input_set.union(op.inputs)
            output_set = output_set.union(op.outputs)
        self.name = "group"
        self.inputs = list(input_set)
        self.outputs = list(output_set)

    def __repr__(self) -> str:
        return (
            f"{self.name}(inputs={self.inputs}, ouputs={self.outputs}, ops={self.ops})"
        )

    def __str__(self) -> str:
        return f"""
            {self.name}
            (
                inputs={self.inputs}, 
                ouputs={self.outputs}, 
                ops={self.ops}
            )
            """


class GraphNode(DataNode):
    def __init__(
        self,
        name: str = "",
        location: DataLocation = DataLocation.CLIENT,
        src: Operator | None = None,
    ) -> None:
        super().__init__(name, location)
        self.src = src


class InputGraphNode(GraphNode):
    def __init__(
        self,
        name: str = "",
        location: DataLocation = DataLocation.CLIENT,
        src: Operator | None = None,
        format: DataFormat | None = None,
        filename: str | None = None,
        default: Any | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(name, location, src)
        self.format = format
        self.filename = filename
        self.default = default
        self.kwargs = kwargs


class MapOperator(Operator, metaclass=abc.ABCMeta):
    @property
    def type(self) -> OpType:
        return OpType.MAP

    @abc.abstractmethod
    def map(self, *args, **kwargs) -> Any:
        ...


class MapOperatorGroup(OperatorGroup):
    def __init__(self, *ops: MapOperator) -> None:
        assert all(op.type == OpType.MAP for op in ops)
        super().__init__(*ops)
        self.name = "map_group"

    @property
    def type(self) -> OpType:
        return OpType.MAP


class ReduceOperator(Operator, metaclass=abc.ABCMeta):
    @property
    def type(self) -> OpType:
        return OpType.REDUCE

    @abc.abstractmethod
    def reduce(self, *args, **kwargs) -> Any:
        ...


class ReduceOperatorGroup(OperatorGroup):
    def __init__(self, *ops: ReduceOperator) -> None:
        assert all(op.type == OpType.REDUCE for op in ops)
        super().__init__(*ops)
        self.name = "reduce_group"

    @property
    def type(self) -> OpType:
        return OpType.REDUCE


AggValueType = Union[np.ndarray, pandas.DataFrame, pandas.Series]

AggResultType = Dict[str, AggValueType]


class MapReduceOperator(Operator, metaclass=abc.ABCMeta):
    def __init__(
        self,
        name: str,
        map_inputs: List[DataNode],
        reduce_inputs: List[DataNode],
        map_outputs: List[DataNode],
        reduce_outputs: List[DataNode],
        **kwargs: Any,
    ) -> None:
        inputs = list(set(map_inputs) | set(reduce_inputs))
        outputs = list(set(map_outputs) | set(reduce_outputs))
        super().__init__(name, inputs, outputs, **kwargs)
        self.map_inputs = map_inputs
        self.reduce_inputs = reduce_inputs
        self.map_outputs = map_outputs
        self.reduce_outputs = reduce_outputs

    @property
    def agg_name(self) -> str:
        return "agg({})".format(",".join(var.name for var in self.reduce_outputs))

    @property
    def type(self) -> OpType:
        return OpType.MAP_REDUCE

    @abc.abstractmethod
    def map(self, *args, **kwargs) -> AggResultType | Tuple[AggResultType, ...]:
        ...

    @abc.abstractmethod
    def reduce(self, data: AggResultType, node_count: int, *args: Any) -> Any:
        ...


class MapReduceOperatorGroup(OperatorGroup):
    def __init__(self, *ops: MapReduceOperator) -> None:
        assert all(op.type == OpType.MAP_REDUCE for op in ops)
        super().__init__(*ops)
        self.name = "map_reduce_group"
        self.ops = ops

    @property
    def agg_names(self) -> List[str]:
        return [op.agg_name for op in self.ops]

    @property
    def type(self) -> OpType:
        return OpType.MAP_REDUCE


class Context(abc.ABC):
    @abc.abstractmethod
    def get(self, *vars: DataNode) -> List[Any]:
        ...

    @abc.abstractmethod
    def set(self, *pairs: Tuple[DataNode, Any]):
        ...

    @abc.abstractmethod
    def has(self, var: DataNode) -> bool:
        ...

    @abc.abstractmethod
    def execute(self, group: OperatorGroup):
        ...

    @abc.abstractmethod
    def clear(self):
        ...


def _get_inputs(ctx: Context, inputs: List[DataNode]) -> List[Any]:
    return ctx.get(*inputs)


def _set_outputs(ctx: Context, outputs: List[DataNode], result: Any):
    if len(outputs) == 1:
        result = (result,)
    assert len(outputs) == len(result), "result length is not equal to outputs length"
    ctx.set(*zip(outputs, result))


class ClientContext(Context):
    @abc.abstractmethod
    def upload(self, name: str, data: Dict[str, Any]):
        ...

    def _execute_op(self, op: Operator):
        if isinstance(op, ReduceOperator):
            raise TypeError("cannot execute reduce operator on client side")

        elif isinstance(op, MapOperator):
            args = _get_inputs(self, op.inputs)
            res = op.map(*args)
            _set_outputs(self, op.outputs, res)

        elif isinstance(op, MapReduceOperator):
            args = _get_inputs(self, op.map_inputs)
            res = op.map(*args)
            if isinstance(res, tuple):
                agg_result = res[0]
                vars = res[1:]
                _set_outputs(self, op.map_outputs, vars)
            else:
                agg_result = res
            self.upload(op.agg_name, agg_result)

    def execute(self, group: OperatorGroup):
        if isinstance(group, ReduceOperatorGroup):
            raise TypeError("cannot execute reduce operator on client side")
        for op in group.ops:
            self._execute_op(op)


class ServerContext(Context):
    @abc.abstractmethod
    def gather(self, name: str) -> Tuple[Dict[str, Any], int]:
        ...

    def _execute_op(self, op: Operator):
        if isinstance(op, MapOperator):
            raise TypeError("cannot execute map operator on server side")

        elif isinstance(op, MapReduceOperator):
            data, node_count = self.gather(op.agg_name)
            args = _get_inputs(self, op.reduce_inputs)
            res = op.reduce(data, node_count, *args)
            _set_outputs(self, op.reduce_outputs, res)

        elif isinstance(op, ReduceOperator):
            args = _get_inputs(self, op.inputs)
            res = op.reduce(*args)
            _set_outputs(self, op.outputs, res)

    def execute(self, group: OperatorGroup):
        if isinstance(group, MapOperatorGroup):
            raise TypeError("cannot execute map operator on server side")
        for op in group.ops:
            self._execute_op(op)


@dataclass
class Step(object):
    map_ops: List[MapOperatorGroup] = field(default_factory=list)
    map_reduce_ops: List[MapReduceOperatorGroup] = field(default_factory=list)
    reduce_ops: List[ReduceOperatorGroup] = field(default_factory=list)

    _inputs: List[DataNode] | None = field(init=False, default=None)
    _outputs: List[DataNode] | None = field(init=False, default=None)
    _agg_names: List[str] | None = field(init=False, default=None)

    @property
    def inputs(self) -> List[DataNode]:
        if self._inputs is None:
            inputs = set()
            for layer in [self.map_ops, self.map_reduce_ops, self.reduce_ops]:
                for group in layer:
                    inputs = inputs.union(group.inputs)
            outputs = set()
            for layer in [self.map_ops, self.map_reduce_ops, self.reduce_ops]:
                for group in layer:
                    outputs = outputs.union(group.outputs)
            self._inputs = list(inputs - outputs)
            return self._inputs
        else:
            return self._inputs

    @property
    def outputs(self) -> List[DataNode]:
        if self._outputs is None:
            inputs = set()
            for layer in [self.map_ops, self.map_reduce_ops, self.reduce_ops]:
                for group in layer:
                    inputs = inputs.union(group.inputs)
            outputs = set()
            for layer in [self.map_ops, self.map_reduce_ops, self.reduce_ops]:
                for group in layer:
                    outputs = outputs.union(group.outputs)
            self._outputs = list(outputs - inputs)
            return self._outputs
        else:
            return self._outputs

    @property
    def agg_names(self) -> List[str]:
        if self._agg_names is None:
            self._agg_names = []
            for op_group in self.map_reduce_ops:
                self._agg_names.extend(op_group.agg_names)
            return self._agg_names
        else:
            return self._agg_names

    def map(self, ctx: ClientContext):
        # map
        for group in self.map_ops:
            ctx.execute(group)

        # map of map reduce
        for group in self.map_reduce_ops:
            ctx.execute(group)

    def reduce(self, ctx: ServerContext):
        # reduce of map reduce
        for group in self.map_reduce_ops:
            ctx.execute(group)

        # reduce
        for group in self.reduce_ops:
            ctx.execute(group)


class TaskType(str, Enum):
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"


@dataclass
class Task(object):
    name: str
    dataset: List[str]
    type: TaskType
    steps: List[Step]
    inputs: List[InputGraphNode]
    outputs: List[DataNode]
    strategy: Strategy


@dataclass
class TaskConstructer(object):
    name: str
    inputs: List[InputGraphNode]
    outputs: List[GraphNode]
    strategy: Strategy
    type: TaskType


def build(constructor: TaskConstructer) -> Task:
    graph = nx.DiGraph()
    # bfs build graph
    queue: List[GraphNode | Operator] = list(constructor.outputs)
    while len(queue):
        next_queue = []
        for node in queue:
            if isinstance(node, GraphNode):
                if node.src:
                    graph.add_edge(node.src, node)
                    next_queue.append(node.src)
            elif isinstance(node, Operator):
                for src in node.inputs:
                    graph.add_edge(src, node)
                    next_queue.append(src)
        queue = next_queue

    # name variables
    var_count = 0
    for node in nx.topological_sort(graph):
        if isinstance(node, GraphNode):
            if len(node.name) == 0:
                node.name = f"_{var_count}"
            var_count += 1

    # sort operators
    op_layers = []
    inputs = constructor.inputs
    init_ready = False
    for layer in nx.topological_generations(graph):
        if isinstance(layer[0], Operator):
            op_layers.append(layer)
        elif not init_ready and isinstance(layer[0], GraphNode):
            assert all(
                isinstance(node, InputGraphNode) for node in layer
            ), "All input variables should be instance of InputGraphNode"
            assert (
                len(set(inputs) - set(layer)) == 0
            ), "Graph first layer doesn't include all inputs"
            inputs = layer
            init_ready = True

    dataset = []
    for node in inputs:
        if (
            isinstance(node, InputGraphNode)
            and node.filename is not None
            and node.format is not None
        ):
            dataset.append(node.filename)

    # construct operator matrix
    step_layers: List[
        Tuple[
            MapOperatorGroup | None,
            MapReduceOperatorGroup | None,
            ReduceOperatorGroup | None,
        ]
    ] = []
    for layer in op_layers:
        map_ops: List[MapOperator] = []
        map_reduce_ops: List[MapReduceOperator] = []
        reduce_ops: List[ReduceOperator] = []
        for op in layer:
            if isinstance(op, MapOperator):
                map_ops.append(op)
            elif isinstance(op, MapReduceOperator):
                map_reduce_ops.append(op)
            elif isinstance(op, ReduceOperator):
                reduce_ops.append(op)
        if len(map_ops) > 0:
            map_group = MapOperatorGroup(*map_ops)
        else:
            map_group = None
        if len(map_reduce_ops) > 0:
            map_reduce_group = MapReduceOperatorGroup(*map_reduce_ops)
        else:
            map_reduce_group = None
        if len(reduce_ops) > 0:
            reduce_group = ReduceOperatorGroup(*reduce_ops)
        else:
            reduce_group = None
        step_layers.append((map_group, map_reduce_group, reduce_group))

    # greedy algorithm to generate steps
    ready_vars = set(inputs)
    steps = []
    map_i, map_reduce_i, reduce_i = 0, 0, 0
    layer_count = len(step_layers)
    while map_i < layer_count or map_reduce_i < layer_count or reduce_i < layer_count:
        step = Step()
        while map_i < layer_count:
            op = step_layers[map_i][0]
            if op is None:
                map_i += 1
            elif len(set(op.inputs) & ready_vars) == len(op.inputs):
                step.map_ops.append(op)
                ready_vars = ready_vars.union(op.outputs)
                map_i += 1
            else:
                break
        map_reduce_outputs = []
        while map_reduce_i < layer_count:
            op = step_layers[map_reduce_i][1]
            if op is None:
                map_reduce_i += 1
            elif len(set(op.inputs) & ready_vars) == len(op.inputs):
                step.map_reduce_ops.append(op)
                map_reduce_outputs.extend(op.outputs)
                map_reduce_i += 1
            else:
                break
        ready_vars = ready_vars.union(map_reduce_outputs)
        while reduce_i < layer_count:
            op = step_layers[reduce_i][2]
            if op is None:
                reduce_i += 1
            elif len(set(op.inputs) & ready_vars) == len(op.inputs):
                step.reduce_ops.append(op)
                ready_vars = ready_vars.union(op.outputs)
                reduce_i += 1
            else:
                break
        steps.append(step)

    task = Task(
        name=constructor.name,
        dataset=dataset,
        type=constructor.type,
        steps=steps,
        inputs=inputs,
        outputs=list(constructor.outputs),
        strategy=constructor.strategy,
    )

    final_step = task.steps[-1]
    if len(final_step.map_reduce_ops) == 0 and len(final_step.reduce_ops) == 0:
        raise ValueError(
            "Should use a map reduce operator or a reduce operator as result"
        )
    if not all(var.location == DataLocation.SERVER for var in task.outputs):
        raise ValueError("All outputs should be on the server")

    return task



from __future__ import annotations

import abc
from typing import Dict, List, Tuple

from ..core.strategy import Strategy
from ..core.task import (
    GraphNode,
    InputGraphNode,
    Task,
    TaskConstructer,
    TaskType,
    build,
)


class HorizontalTask(abc.ABC):
    def __init__(self, name: str, strategy: Strategy) -> None:
        self.name = name
        self.strategy = strategy

    @abc.abstractmethod
    def dataset(self) -> Dict[str, InputGraphNode] | InputGraphNode:
        ...

    @abc.abstractmethod
    def _build_graph(self) -> Tuple[List[InputGraphNode], List[GraphNode]]:
        ...

    def build(self) -> Task:
        inputs, outputs = self._build_graph()
        constructor = TaskConstructer(
            name=self.name,
            inputs=inputs,
            outputs=outputs,
            strategy=self.strategy,
            type=TaskType.HORIZONTAL,
        )
        return build(constructor)



from __future__ import annotations

import errno
import os
import random
from typing import Any, List, Tuple

import pandas as pd
from PIL import Image
from torch.utils.data import Dataset as TorchDataset

from ..core.task import DataFormat, DataLocation, InputGraphNode
from .file import load_file


class Dataset(InputGraphNode):
    def __init__(
        self,
        dataset: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name="",
            location=DataLocation.CLIENT,
            src=None,
            format=DataFormat.DATASET,
            filename=dataset,
            default=None,
            **kwargs,
        )


class FileDataset(TorchDataset):
    def __init__(self, filename: str) -> None:
        result = load_file(filename)
        if isinstance(result, Image.Image):
            raise ValueError("file dataset does not support image file")
        self._result = result

    def __getitem__(self, index):
        if isinstance(self._result, pd.DataFrame):
            item = self._result.iloc[index]
        else:
            item = self._result[index]

        return item

    def __len__(self) -> int:
        return len(self._result)


class DirectoryDataset(TorchDataset):
    def __init__(self, directory: str) -> None:
        self._xs = []
        self._ys = []
        root, dirnames, filenames = next(os.walk(directory))
        if len(filenames) > 0 and len(dirnames) == 0:
            self._xs.extend([os.path.join(root, filename) for filename in filenames])
        elif len(filenames) == 0 and len(dirnames) > 0:
            for dirname in dirnames:
                sub_root, sub_dirs, sub_files = next(
                    os.walk(os.path.join(root, dirname))
                )
                if len(sub_files) > 0 and len(sub_dirs) == 0:
                    self._xs.extend(
                        [os.path.join(sub_root, filename) for filename in sub_files]
                    )
                    self._ys.extend([dirname] * len(sub_files))
                else:
                    raise ValueError(
                        "directory can only contain files or sub directory that contains file"
                    )
        else:
            raise ValueError(
                "directory can only contain files or sub directory that contains file"
            )

    def __getitem__(self, index):
        filename = self._xs[index]
        x = load_file(filename)
        y = None
        if len(self._ys) > 0:
            y = self._ys[index]
        if y is not None:
            return x, y
        return x

    def __len__(self) -> int:
        return len(self._xs)


class IndexLookUpDataset(TorchDataset):
    def __init__(self, indices: List[int], dataset: TorchDataset) -> None:
        self._indices = indices
        self._dataset = dataset

    def __getitem__(self, index):
        i = self._indices[index]
        return self._dataset[i]

    def __len__(self) -> int:
        return len(self._indices)


def split_dataset(
    dataset: TorchDataset, frac: float, seed: int = 0
) -> Tuple[IndexLookUpDataset, IndexLookUpDataset]:
    assert frac > 0 and frac < 1

    size = len(dataset)  # type: ignore
    frac_count = int(size * frac)
    indices = list(range(size))

    state = random.Random(seed)
    state.shuffle(indices)

    left_indices, right_indices = indices[:frac_count], indices[frac_count:]
    return IndexLookUpDataset(left_indices, dataset), IndexLookUpDataset(
        right_indices, dataset
    )


def load_dataset(
    dataset_name: str,
) -> TorchDataset | Tuple[TorchDataset, TorchDataset]:
    if not os.path.exists(dataset_name):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), dataset_name)
    if os.path.isfile(dataset_name):
        dataset = FileDataset(dataset_name)
        return dataset
    else:
        train_path = os.path.join(dataset_name, "train")
        val_path = os.path.join(dataset_name, "val")
        if (
            os.path.exists(train_path)
            and os.path.isdir(train_path)
            and os.path.exists(val_path)
            and os.path.isdir(val_path)
        ):
            train_root, _, train_files = next(os.walk(train_path))
            val_root, _, val_files = next(os.walk(val_path))
            if len(train_files) == 1 and len(val_files) == 1:
                train_dataset = FileDataset(os.path.join(train_root, train_files[0]))
                val_dataset = FileDataset(os.path.join(val_root, val_files[0]))
            else:
                train_dataset = DirectoryDataset(train_path)
                val_dataset = DirectoryDataset(val_path)
            return train_dataset, val_dataset
        else:
            dataset = DirectoryDataset(dataset_name)
            return dataset



# type: ignore
# -*- coding: utf-8 -*-
import shlex
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
        self.pytest_args = ["tests"] + shlex.split(self.pytest_args)

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name="delta-task",
    version="0.5.1",
    license_files=("LICENSE"),
    packages=find_packages(),
    include_package_data=True,
    exclude_package_data={"": [".gitignore"]},
    install_requires=[
        "cloudpickle==1.6.0",
        "httpx==0.21.1",
        "numpy==1.21.4",
        "Pillow==8.4.0",
        "pandas==1.2.3",
        "pytest==6.2.5",
        "torch==1.8.2+cpu",
        "networkx==2.7.1"
    ],
    tests_require=["pytest"],
    cmdclass={"test": PyTest},
    test_suite="tests",
    zip_safe=False,
    author="miaohong",
    author_email="73902525@qq.com",
    description="delta framework",
    python_requires=">=3.6",
    url="https://github.com/delta-mpc/delta",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
)



from __future__ import annotations

import abc
import logging
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Type

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
import delta.dataset

from ..dataset import split_dataset
from ..core.strategy import (
    LearningStrategy,
    RandomSelectStrategy,
    EpochMergeStrategy,
    IterMergeStrategy,
    WeightResultStrategy,
    CURVE_TYPE,
    ResultStrategy,
    SelectStrategy,
)
from ..core.task import (
    DataFormat,
    DataLocation,
    DataNode,
    GraphNode,
    InputGraphNode,
    MapOperator,
    MapReduceOperator,
    Operator,
    ReduceOperator,
)
from .task import HorizontalTask

_logger = logging.getLogger(__name__)


class FedAvg(LearningStrategy):
    def __init__(
        self,
        min_clients: int = 2,
        max_clients: int = 2,
        merge_epoch: int = 1,
        merge_iteration: int = 0,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
        select_strategy: Type[SelectStrategy] = RandomSelectStrategy,
        result_strategy: Type[ResultStrategy] = WeightResultStrategy,
    ) -> None:
        if merge_epoch > 0 and merge_iteration > 0:
            raise ValueError(
                "merge epoch and merge iteration cannot be both greater than 0"
            )
        if merge_epoch > 0:
            merge_strategy = EpochMergeStrategy(merge_epoch)
        else:
            merge_strategy = IterMergeStrategy(merge_iteration)
        super().__init__(
            "FedAvg",
            select_strategy(min_clients, max_clients),
            merge_strategy,
            result_strategy(),
            wait_timeout,
            connection_timeout,
            False,
            precision,
            curve,
        )


class FaultTolerantFedAvg(LearningStrategy):
    def __init__(
        self,
        min_clients: int = 2,
        max_clients: int = 2,
        merge_epoch: int = 0,
        merge_iteration: int = 0,
        wait_timeout: float = 60,
        connection_timeout: float = 60,
        precision: int = 8,
        curve: CURVE_TYPE = "secp256k1",
        select_strategy: Type[SelectStrategy] = RandomSelectStrategy,
        result_strategy: Type[ResultStrategy] = WeightResultStrategy,
    ) -> None:
        if merge_epoch > 0 and merge_iteration > 0:
            raise ValueError(
                "merge epoch and merge iteration cannot be both greater than 0"
            )
        if merge_epoch <= 0 and merge_iteration <= 0:
            raise ValueError(
                "one of merge epoch and merge iteration should be greater than 0"
            )
        if merge_epoch > 0:
            merge_strategy = EpochMergeStrategy(merge_epoch)
        else:
            merge_strategy = IterMergeStrategy(merge_iteration)

        super().__init__(
            "FaultTolerantFedAvg",
            select_strategy(min_clients, max_clients),
            merge_strategy,
            result_strategy(),
            wait_timeout,
            connection_timeout,
            True,
            precision,
            curve,
        )


class TrainIterator(object):
    def __init__(
        self,
        dataloader: DataLoader,
        epoch: int,
        iteration: int,
        strategy: LearningStrategy,
    ) -> None:
        self.dataloader = dataloader
        self.epoch = epoch
        self.iteration = iteration
        self.strategy = strategy

    def __iter__(self):
        return self._get_iter()

    def _get_iter(self):
        finished = False

        while not finished:
            for batch in self.dataloader:
                if finished:
                    break

                _logger.info(f"Training epoch {self.epoch} iteration {self.iteration}")

                yield batch

                if self.strategy.should_merge(self.epoch, self.iteration, False):
                    _logger.info(f"iteration {self.iteration}, start to merge")
                    finished = True
                self.iteration += 1

            if self.strategy.should_merge(self.epoch, self.iteration, True):
                _logger.info(f"epoch {self.epoch}, start to merge")
                finished = True
            if (self.iteration - 1) % len(self.dataloader) == 0:
                self.epoch += 1


class ValidateIterator(object):
    def __init__(self, dataloader: DataLoader) -> None:
        self.dataloader = dataloader

    def __iter__(self):
        return self._get_iter()

    def _get_iter(self):
        iteration = 1
        for batch in self.dataloader:
            _logger.info(f"Validation iteration {iteration}")
            yield batch
            iteration += 1


class HorizontalLearning(HorizontalTask):
    def __init__(
        self,
        name: str,
        max_rounds: int,
        validate_interval: int = 1,
        validate_frac: float = 0.2,
        strategy: LearningStrategy = FaultTolerantFedAvg(
            merge_epoch=1,
        ),
    ) -> None:
        super().__init__(name=name, strategy=strategy)
        self.strategy = strategy
        self.max_rounds = max_rounds

        self.validate_interval = validate_interval
        self.validate_frac = validate_frac
        self.enable_validate = validate_interval > 0

    @abc.abstractmethod
    def make_train_dataloader(self, dataset: Dataset) -> DataLoader:
        ...

    @abc.abstractmethod
    def make_validate_dataloader(self, dataset: Dataset) -> DataLoader:
        ...

    @abc.abstractmethod
    def train(self, dataloader: Iterable):
        ...

    @abc.abstractmethod
    def validate(self, dataloader: Iterable) -> Dict[str, Any]:
        ...

    @abc.abstractmethod
    def state_dict(self) -> Dict[str, torch.Tensor]:
        ...

    @abc.abstractmethod
    def dataset(self) -> delta.dataset.Dataset:
        ...

    def _dataset_nodes(
        self, dataset_node: delta.dataset.Dataset
    ) -> Tuple[GraphNode] | Tuple[GraphNode, GraphNode]:
        # train_dataset = DatasetNode(
        #     name="train_dataset",
        #     location=DataLocation.CLIENT,
        #     dataset=self.dataset,
        #     validate=False,
        #     validate_frac=self.validate_frac,
        #     seed=self.seed,
        # )
        # if self.enable_validate:
        #     validate_dataset = DatasetNode(
        #         name="validate_dataset",
        #         location=DataLocation.CLIENT,
        #         dataset=self.dataset,
        #         validate=True,
        #         validate_frac=self.validate_frac,
        #         seed=self.seed,
        #     )
        #     return train_dataset, validate_dataset
        # else:
        #     return (train_dataset,)
        class _SplitDatasetOp(MapOperator):
            def __init__(
                self,
                name: str,
                inputs: List[DataNode],
                outputs: List[DataNode],
                validate: bool,
                validate_frac: float,
            ) -> None:
                super().__init__(
                    name,
                    inputs,
                    outputs,
                    validate=validate,
                    validate_frac=validate_frac,
                )
                self.validate = validate
                self.validate_frac = validate_frac

            def map(
                self, dataset: Dataset | Tuple[Dataset, Dataset]
            ) -> Dataset | Tuple[Dataset, Dataset]:
                if isinstance(dataset, tuple):
                    train_dataset, val_dataset = dataset
                    if self.validate:
                        return train_dataset, val_dataset
                    else:
                        return train_dataset
                else:
                    if self.validate:
                        train_dataset, val_dataset = split_dataset(
                            dataset, self.validate_frac
                        )
                        return train_dataset, val_dataset
                    else:
                        return dataset

        if self.enable_validate:
            train_dataset = GraphNode(
                name="train_dataset", location=DataLocation.CLIENT
            )
            val_dataset = GraphNode(name="val_dataset", location=DataLocation.CLIENT)
            outputs = (train_dataset, val_dataset)
        else:
            train_dataset = GraphNode(
                name="train_dataset", location=DataLocation.CLIENT
            )
            outputs = (train_dataset,)

        split_op = _SplitDatasetOp(
            name="split_dataset",
            inputs=[dataset_node],
            outputs=list(outputs),
            validate=self.enable_validate,
            validate_frac=self.validate_frac,
        )
        for node in outputs:
            node.src = split_op

        return outputs

    def _dataloader_nodes(
        self,
        train_dataset_node: GraphNode,
        val_dataset_node: Optional[GraphNode] = None,
    ) -> Tuple[GraphNode, ...]:
        train_dataloader_node = GraphNode(
            name=f"train_dataloader", location=DataLocation.CLIENT
        )

        class _DataloaderOp(MapOperator):
            def __init__(
                self,
                name: str,
                inputs: List[DataNode],
                outputs: List[DataNode],
                learning: "HorizontalLearning",
                train: bool,
            ) -> None:
                super().__init__(name, inputs, outputs, learning=learning, train=train)
                self.learning = learning
                self.train = train

            def map(self, dataset: Dataset) -> Iterable:
                if self.train:
                    dataloader = self.learning.make_train_dataloader(dataset)
                else:
                    dataloader = self.learning.make_validate_dataloader(dataset)
                return dataloader

        train_dataloader_op = _DataloaderOp(
            name="make_train_dataloader",
            inputs=[train_dataset_node],
            outputs=[train_dataloader_node],
            learning=self,
            train=True,
        )
        train_dataloader_node.src = train_dataloader_op

        res = [train_dataloader_node]

        if val_dataset_node is not None:
            val_dataloader_node = GraphNode(
                name=f"validate_dataloader", location=DataLocation.CLIENT
            )

            val_dataloader_op = _DataloaderOp(
                name="make_validate_dataloader",
                inputs=[val_dataset_node],
                outputs=[val_dataloader_node],
                learning=self,
                train=False,
            )
            val_dataloader_node.src = val_dataloader_op
            res.append(val_dataloader_node)

        return tuple(res)

    def _train_result_node(
        self,
        train_dataloader_node: GraphNode,
        weight_node: GraphNode,
        epoch_node: GraphNode,
        iteration_node: GraphNode,
        round: int,
    ) -> GraphNode:
        new_weight_node = GraphNode(
            name=f"weight_{round}", location=DataLocation.SERVER
        )
        new_epoch_node = GraphNode(name="epoch", location=DataLocation.CLIENT)
        new_iteration_node = GraphNode(name="iteration", location=DataLocation.CLIENT)

        class _TrainOp(MapReduceOperator):
            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                learning: "HorizontalLearning",
                round: int,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    learning=learning,
                    round=round,
                )
                self.learning = learning
                self.round = round

            def map(
                self,
                dataloader: DataLoader,
                weight: np.ndarray,
                epoch: int,
                iteration: int,
            ) -> Tuple[Dict[str, np.ndarray], int, int]:
                self.learning.strategy.weight_to_params(
                    weight, self.learning.state_dict()
                )
                _logger.info(f"Round {self.round} training")
                train_iter = TrainIterator(
                    dataloader, epoch, iteration, self.learning.strategy
                )
                self.learning.train(train_iter)
                _logger.info(f"Round {self.round} training complete")
                params = self.learning.state_dict()
                res = self.learning.strategy.params_to_result(params, weight)
                return {"res": res}, train_iter.epoch, train_iter.iteration

            def reduce(
                self, data: Dict[str, np.ndarray], node_count: int, weight: np.ndarray
            ) -> np.ndarray:
                res = data.get("res")
                if res is not None:
                    res = res / node_count
                    params = self.learning.state_dict()
                    self.learning.strategy.weight_to_params(weight, params)
                    self.learning.strategy.result_to_params(res, params)
                    new_weight = self.learning.strategy.params_to_weight(params)
                    return new_weight
                else:
                    raise ValueError("aggregate result missing key 'res'")

        train_op = _TrainOp(
            name="train",
            map_inputs=[train_dataloader_node, weight_node, epoch_node, iteration_node],
            reduce_inputs=[weight_node],
            map_outputs=[new_epoch_node, new_iteration_node],
            reduce_outputs=[new_weight_node],
            learning=self,
            round=round,
        )
        new_weight_node.src = train_op

        return new_weight_node

    def _val_result_node(
        self,
        val_dataloader_node: GraphNode,
        weight_node: GraphNode,
        round: int,
        metrics_node: GraphNode | None = None,
    ) -> GraphNode:
        """
        val_result_node: No use. Adding this node can make the dependency more straight forward,
        and simplify the final ouputs nodes.
        """
        new_metrics_node = GraphNode(
            name=f"metrics_{round}", location=DataLocation.SERVER
        )

        class _ValidateOp(MapReduceOperator):
            priority: int = 1

            def __init__(
                self,
                name: str,
                map_inputs: List[DataNode],
                reduce_inputs: List[DataNode],
                map_outputs: List[DataNode],
                reduce_outputs: List[DataNode],
                learning: "HorizontalLearning",
                round: int,
            ) -> None:
                super().__init__(
                    name,
                    map_inputs,
                    reduce_inputs,
                    map_outputs,
                    reduce_outputs,
                    learning=learning,
                    round=round,
                )
                self.learning = learning
                self.round = round

            def map(
                self, dataloader: DataLoader, weight: np.ndarray
            ) -> Dict[str, np.ndarray]:
                self.learning.strategy.weight_to_params(
                    weight, self.learning.state_dict()
                )
                _logger.info(f"Round {self.round} validating")
                val_iter = ValidateIterator(dataloader)
                metrics = self.learning.validate(val_iter)
                _logger.info(f"Round {self.round} validating complete")
                res = {key: np.array(val) for key, val in metrics.items()}
                return res

            def reduce(
                self, data: Dict[str, np.ndarray], node_count: int, metrics=None
            ) -> Dict[str, np.ndarray | float]:
                res = {}
                for key, val in data.items():
                    tmp = val / node_count
                    try:
                        res[key] = tmp.item()
                    except ValueError:
                        res[key] = tmp
                return res

        val_op = _ValidateOp(
            name="validate",
            map_inputs=[val_dataloader_node, weight_node],
            reduce_inputs=[] if metrics_node is None else [metrics_node],
            map_outputs=[],
            reduce_outputs=[new_metrics_node],
            learning=self,
            round=round,
        )
        new_metrics_node.src = val_op

        return new_metrics_node

    def _result_node(
        self, weight_node: GraphNode, metrics_node: GraphNode | None = None
    ) -> GraphNode:
        result_node = GraphNode(name="result", location=DataLocation.SERVER)

        class _ResultOp(ReduceOperator):
            def __init__(
                self,
                name: str,
                inputs: List[DataNode],
                outputs: List[DataNode],
                learning: "HorizontalLearning",
            ) -> None:
                super().__init__(name, inputs, outputs, learning=learning)
                self.learning = learning

            def reduce(
                self,
                weight: np.ndarray,
                metrics: Dict[str, np.ndarray | float] | None = None,
            ) -> Any:
                self.learning.strategy.weight_to_params(
                    weight, self.learning.state_dict()
                )
                return self.learning.state_dict()

        input_nodes: List[DataNode] = [weight_node]
        if metrics_node is not None:
            input_nodes.append(metrics_node)

        result_op = _ResultOp(
            name="result", inputs=input_nodes, outputs=[result_node], learning=self
        )
        result_node.src = result_op
        return result_node

    def _build_graph(self) -> Tuple[List[delta.dataset.Dataset], List[GraphNode]]:
        dataset_node = self.dataset()
        dataset_node.name = "dataset"
        dataset_nodes = self._dataset_nodes(dataset_node)
        dataloader_nodes = self._dataloader_nodes(*dataset_nodes)

        epoch_node = InputGraphNode(
            name="epoch", location=DataLocation.CLIENT, default=1
        )
        iteration_node = InputGraphNode(
            name="iteration", location=DataLocation.CLIENT, default=1
        )
        weight_arr = self.strategy.params_to_weight(self.state_dict())
        weight_node = InputGraphNode(
            name="weight_0", location=DataLocation.SERVER, default=weight_arr
        )
        metrics_node = None
        inputs = [dataset_node, epoch_node, iteration_node, weight_node]
        for i in range(self.max_rounds):
            train_dataloader_node = dataloader_nodes[0]
            if len(dataloader_nodes) > 1:
                val_dataloader_node = dataloader_nodes[1]
            else:
                val_dataloader_node = None

            weight_node = self._train_result_node(
                train_dataloader_node, weight_node, epoch_node, iteration_node, i + 1
            )
            if (
                val_dataloader_node is not None
                and (i + 1) % self.validate_interval == 0
            ):
                metrics_node = self._val_result_node(
                    val_dataloader_node, weight_node, i + 1, metrics_node
                )

        output_node = self._result_node(weight_node, metrics_node)

        return inputs, [output_node]
